/*
 * -----------------------------------------------------------------------------
 *
 * <p><b>License and Copyright: </b>The contents of this file are subject to the
 * Apache License, Version 2.0 (the "License"); you may not use 
 * this file except in compliance with the License. You may obtain a copy of 
 * the License at <a href="http://www.fedora-commons.org/licenses">
 * http://www.fedora-commons.org/licenses.</a></p>
 *
 * <p>Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.</p>
 *
 * <p>The entire file consists of original code.</p>
 * <p>Copyright &copy; 2008 Fedora Commons, Inc.<br />
 * <p>Copyright &copy; 2002-2007 The Rector and Visitors of the University of 
 * Virginia and Cornell University<br /> 
 * All rights reserved.</p>
 *
 * -----------------------------------------------------------------------------
 */
package fedora.server.resourceIndex;

import java.util.TreeSet;

/**
 * A sorted set of domain values for a parameter.
 *
 * As per the <code>SortedSet</code> contract, iterators over the values
 * in this collection will provide the elements in ascending order.
 *
 * @author cwilper@cs.cornell.edu
 */
public class ParamDomain extends TreeSet<String> {
	private static final long serialVersionUID = 1L;

    /**
     * The parameter whose domain is being described.
     */
    private String _parameterName;

    /**
     * Whether specifying a value is required.
     */
    private boolean _isRequired;

    /**
     * Construct an empty <code>ParamDomain</code>.
     *
     * @param parameterName the parameter whose domain is being described.
     * @param isRequired whether specifying a value is required.
     */
    public ParamDomain(String parameterName, boolean isRequired) {
        _parameterName = parameterName;
        _isRequired = isRequired;
    }

    /**
     * Construct a <code>ParamDomain</code> with values from the given array.
     *
     * @param parameterName the parameter whose domain is being described.
     * @param isRequired whether specifying a value is required.
     * @param values the domain values.
     */
    public ParamDomain(String parameterName, boolean isRequired, 
            String[] domainValues) {
        _parameterName = parameterName;
        _isRequired = isRequired;
        for (int i = 0; i < domainValues.length; i++) {
            add(domainValues[i]);
        }
    }

    /**
     * Get the name of the parameter whose domain is being described.
     */
    public String getParameterName() {
        return _parameterName;
    }

    /**
     * Tell whether specifying a value is required.
     */
    public boolean isRequired() {
        return _isRequired;
    }

    public boolean equals(Object obj) {
        if (super.equals(obj)) {
            ParamDomain p = (ParamDomain) obj;
            return _parameterName.equals(p.getParameterName())
                    && _isRequired == p.isRequired();
        } else {
            return false;
        }
    }

}
